@@ -23,6 +23,7 @@ import chess
 import collections
 import signal
 import subprocess
+import logging
 import threading
 
 try:
@@ -31,16 +32,26 @@ except ImportError:
     import Queue as queue
 
 
+LOGGER = logging.getLogger(__name__)
+
+
 POLL_TIMEOUT = 5
 STOP_TIMEOUT = 2
 
 
 class Option(collections.namedtuple("Option", ["name", "type", "default", "min", "max", "var"])):
     """Information about an available option for an UCI engine."""
+    pass
 
 
 class Score(collections.namedtuple("Score", ["cp", "mate", "lowerbound", "upperbound"])):
     """A centipawns or mate score sent by an UCI engine."""
+    pass
+
+
+class BestMove(collections.namedtuple("BestMove", ["bestmove", "ponder"])):
+    """A bestmove and ponder move sent by an UCI engine."""
+    pass
 
 
 class OptionMap(collections.MutableMapping):
@@ -241,98 +252,122 @@ class InfoHandler(object):
         self.release()
 
 
-class CommandTimeoutException(Exception):
+class TimeoutError(Exception):
+    """The UCI command timed out."""
     pass
 
 
 class Command(object):
     """Information about the state of a command."""
+    def __init__(self):
+        self._condition = threading.Condition()
+        self._result = None
+        self._done = False
+        self._done_callbacks = []
 
-    def __init__(self, callback=None):
-        self.result = None
-
-        self._callback = callback
-        self._event = threading.Event()
-
-    def _notify(self, result):
-        self.result = result
-        self._event.set()
-        self._notify_callback()
-
-    def _notify_callback(self):
-        if self._callback and self._callback is not True:
-            if self.result is None:
-                self._callback()
-                return
-
+    def _invoke_callbacks(self):
+        for callback in self._done_callbacks:
             try:
-                iter(self.result)
-            except TypeError:
-                self._callback(self.result)
-                return
-            self._callback(*self.result)
+                callback(self)
+            except Exception:
+                LOGGER.exception("exception calling callback for %r", self)
 
-    def _wait_or_callback(self):
-        if not self._callback:
-            return self.wait()
-        else:
-            return self
+    def __repr__(self):
+        with self._condition:
+            if self._done:
+                if self._result is None:
+                    return "<Command at {0} (finished)>".format(hex(id(self)))
+                else:
+                    return "<Command at {0} (result={1})>".format(hex(id(self)), self._result)
+            else:
+                return "<Command at {0} (pending)>".format(hex(id(self)))
 
-    def wait(self, timeout=None):
+    def done(self):
+        """Returns whether the command has already been completed."""
+        with self._condition:
+            return self._done
+
+    def add_done_callback(self, fn):
         """
-        Wait for the command to be completed.
+        Add a callback function to be notified once the command completes.
 
-        This may wait forever unless a floating point number of seconds is
-        specified as the timeout. Raises *CommandTimeoutException* if a timeout
-        is indeed encountered.
+        The callback function will receive the *Command* object as a single
+        argument.
+
+        The callback might be executed on a different thread. If the command
+        has already been completed it will be invoked immidiately, instead.
         """
-        if not self._event.wait(timeout):
-            raise CommandTimeoutException("waiting for an uci command timed out")
-        return self.result
+        with self._condition:
+            if self._done:
+                fn(self)
+            else:
+                self._done_callbacks.append(fn)
 
-    def is_done(self):
-        """Checks if the command has been completed yet."""
-        return self._event.is_set()
+    def result(self, timeout=None):
+        """
+        Wait for the command to finish and return the result.
+
+        A *timeout* in seconds may be given as a floating point number and
+        *TimeoutError* is raised if the command does not complete in time.
+        """
+        with self._condition:
+            if self._done:
+                return self._result
+
+            self._condition.wait(timeout)
+
+            if self._done:
+                return self._result
+            else:
+                raise TimeoutError()
+
+    def set_result(self, result):
+        with self._condition:
+            self._result = result
+            self._done = True
+            self._condition.notify_all()
+
+        self._invoke_callbacks()
+
+    def execute(self, engine):
+        pass
 
 
 class UciCommand(Command):
-    def __init__(self, callback=None):
-        super(UciCommand, self).__init__(callback)
-
-    def _execute(self, engine):
+    def execute(self, engine):
         engine.uciok.clear()
         engine.send_line("uci")
         engine.uciok.wait()
-        self._notify(None)
+        self.set_result(None)
 
 
 class DebugCommand(Command):
-    def __init__(self, on, callback=None):
-        super(DebugCommand, self).__init__(callback)
-        self.on = on
+    def __init__(self, on):
+        super(DebugCommand, self).__init__()
+        self.on = bool(on)
 
-    def _execute(self, engine):
+    def execute(self, engine):
         if self.on:
             engine.send_line("debug on")
         else:
             engine.send_line("debug off")
-        self._notify(None)
+        self.set_result(None)
 
 
 class IsReadyCommand(Command):
-    def __init__(self, callback=None):
-        super(IsReadyCommand, self).__init__(callback)
+    def __init__(self):
+        super(IsReadyCommand, self).__init__()
 
-    def _execute(self, engine):
+    def execute(self, engine):
         engine.readyok.clear()
         engine.send_line("isready")
         engine.readyok.wait()
-        self._notify(None)
+        self.set_result(None)
 
 
 class SetOptionCommand(IsReadyCommand):
-    def __init__(self, options, callback=None):
-        super(SetOptionCommand, self).__init__(callback)
+    def __init__(self, options):
+        super(SetOptionCommand, self).__init__()
 
         self.option_lines = []
 
@@ -352,25 +387,22 @@ class SetOptionCommand(IsReadyCommand):
 
             self.option_lines.append("".join(builder))
 
-    def _execute(self, engine):
+    def execute(self, engine):
         for option_line in self.option_lines:
             engine.send_line(option_line)
 
-        super(SetOptionCommand, self)._execute(engine)
+        super(SetOptionCommand, self).execute(engine)
 
 
 class UciNewGameCommand(IsReadyCommand):
-    def __init__(self, callback=None):
-        super(UciNewGameCommand, self).__init__(callback)
-
-    def _execute(self, engine):
+    def execute(self, engine):
         engine.send_line("ucinewgame")
-        super(UciNewGameCommand, self)._execute(engine)
+        super(UciNewGameCommand, self).execute(engine)
 
 
 class PositionCommand(IsReadyCommand):
-    def __init__(self, board, callback=None):
-        super(PositionCommand, self).__init__(callback)
+    def __init__(self, board):
+        super(PositionCommand, self).__init__()
 
         builder = []
         builder.append("position")
@@ -396,14 +428,14 @@ class PositionCommand(IsReadyCommand):
 
         self.buf = " ".join(builder)
 
-    def _execute(self, engine):
+    def execute(self, engine):
         engine.send_line(self.buf)
-        super(PositionCommand, self)._execute(engine)
+        super(PositionCommand, self).execute(engine)
 
 
 class GoCommand(Command):
-    def __init__(self, searchmoves=None, ponder=False, wtime=None, btime=None, winc=None, binc=None, movestogo=None, depth=None, nodes=None, mate=None, movetime=None, infinite=False, callback=None):
-        super(GoCommand, self).__init__(callback)
+    def __init__(self, searchmoves=None, ponder=False, wtime=None, btime=None, winc=None, binc=None, movestogo=None, depth=None, nodes=None, mate=None, movetime=None, infinite=False):
+        super(GoCommand, self).__init__()
 
         builder = []
         builder.append("go")
@@ -448,7 +480,6 @@ class GoCommand(Command):
             builder.append("mate")
             builder.append(str(int(mate)))
 
-
         if movetime is not None:
             builder.append("movetime")
             builder.append(str(int(movetime)))
@@ -459,23 +490,20 @@ class GoCommand(Command):
 
         self.buf = " ".join(builder)
 
-    def _execute(self, engine):
+    def execute(self, engine):
         engine.bestmove = None
         engine.ponder = None
         engine.bestmove_received.clear()
         engine.send_line(self.buf)
         if self.infinite:
-            self._notify(None)
+            self.set_result(None)
         else:
             engine.bestmove_received.wait()
-            self._notify((engine.bestmove, engine.ponder))
+            self.set_result(BestMove(engine.bestmove, engine.ponder))
 
 
 class StopCommand(Command):
-    def __init__(self, callback=None):
-        super(StopCommand, self).__init__(callback)
-
-    def _execute(self, engine):
+    def execute(self, engine):
         engine.readyok.clear()
 
         # First check if the engine already sent a best move and stopped
@@ -488,51 +516,47 @@ class StopCommand(Command):
         engine.readyok.wait()
 
         engine.bestmove_received.wait(STOP_TIMEOUT)
-        self._notify((engine.bestmove, engine.ponder))
+        self.set_result(BestMove(engine.bestmove, engine.ponder))
 
 
 class PonderhitCommand(IsReadyCommand):
-    def __init__(self, callback=None):
-        super(PonderhitCommand, self).__init__(callback)
-
-    def _execute(self, engine):
+    def execute(self, engine):
         engine.send_line("ponderhit")
-        super(PonderhitCommand, self)._execute(engine)
+        super(PonderhitCommand, self).execute(engine)
 
 
 class QuitCommand(Command):
-    def __init__(self, engine, callback=None):
-        self.engine = engine
-        self._callback = callback
-
-    def _execute(self, engine):
-        assert self.engine == engine
+    def execute(self, engine):
         engine.send_line("quit")
         engine.terminated.wait()
-        engine.process.close_std_streams()
-        self._notify_callback()
-
-    @property
-    def result(self):
-        return self.engine.return_code
-
-    def wait(self, timeout=None):
-        if not self.engine.terminated.wait(timeout):
-            raise CommandTimeoutException("waiting for engine termination timed out")
-        return self.result
-
-    def is_done(self):
-        return self.engine.terminated.is_set()
+        self.set_result(engine.process.wait_for_return_code())
 
 
-class MockProcess(object):
+class TerminationPromise(object):
     def __init__(self, engine):
         self.engine = engine
 
+    def done(self):
+        return self.engine.terminated.is_set()
+
+    def result(self, timeout=None):
+        self.engine.terminated.wait(timeout)
+
+        if not self.done():
+            raise TimeoutError()
+        else:
+            return self.engine.return_code
+
+
+class MockProcess(object):
+    def __init__(self):
         self._expectations = collections.deque()
         self._is_dead = threading.Event()
         self._std_streams_closed = False
 
+    def spawn(self, engine):
+        self.engine = engine
+
     def expect(self, expectation, responses=()):
         self._expectations.append((expectation, responses))
 
@@ -572,14 +596,23 @@ class MockProcess(object):
         self._is_dead.wait()
         return 0
 
+    def pid(self):
+        return None
+
+    def __repr__(self):
+        return "<MockProcess at {0}>".format(hex(id(self)))
+
 
 class PopenProcess(object):
-    def __init__(self, engine, command):
-        self.engine = engine
-        self.process = subprocess.Popen(command, stdout=subprocess.PIPE, stdin=subprocess.PIPE, bufsize=1, universal_newlines=True)
+    def __init__(self, command):
+        self.command = command
 
         self._receiving_thread = threading.Thread(target=self._receiving_thread_target)
         self._receiving_thread.daemon = True
+
+    def spawn(self, engine):
+        self.engine = engine
+        self.process = subprocess.Popen(self.command, stdout=subprocess.PIPE, stdin=subprocess.PIPE, bufsize=1, universal_newlines=True)
         self._receiving_thread.start()
 
     def _receiving_thread_target(self):
@@ -614,18 +647,29 @@ class PopenProcess(object):
         self.process.wait()
         return self.process.returncode
 
+    def pid(self):
+        return self.process.pid
+
+    def __repr__(self):
+        return "<PopenProcess at {0} (pid={1})>".format(hex(id(self)), self.pid())
+
 
 class SpurProcess(object):
-    def __init__(self, engine, shell, command):
-        self._stdout_buffer = []
+    def __init__(self, shell, command):
+        self.shell = shell
+        self.command = command
 
-        self.engine = engine
-        self.process = shell.spawn(command, store_pid=True, allow_error=True, stdout=self)
+        self._stdout_buffer = []
 
         self._result = None
 
         self._waiting_thread = threading.Thread(target=self._waiting_thread_target)
         self._waiting_thread.daemon = True
+
+    def spawn(self, engine):
+        self.engine = engine
+
+        self.process = self.shell.spawn(self.command, store_pid=True, allow_error=True, stdout=self)
         self._waiting_thread.start()
 
     def write(self, byte):
@@ -652,7 +696,13 @@ class SpurProcess(object):
         self.process.send_signal(signal.SIGKILL)
 
     def close_std_streams(self):
-        # Spur already handles cleanup.
+        # TODO: Spur does not do real clean up.
+        #try:
+        #    self.process._process_stdin.close()
+        #except AttributeError:
+        #    self.process._stdin.close()
+        #self.process._io._handlers[0]._file_in.close()
+        #self.process._io._handlers[1]._file_in.close()
         pass
 
     def send_line(self, string):
@@ -662,10 +712,16 @@ class SpurProcess(object):
     def wait_for_return_code(self):
         return self.process.wait_for_result().return_code
 
+    def pid(self):
+        return self.process.pid
+
+    def __repr__(self):
+        return "<SpurProcess at {0} (pid={1})>".format(hex(id(self)), self.pid())
+
 
 class Engine(object):
-    def __init__(self, process_cls, args):
-        self.process = process_cls(self, *args)
+    def __init__(self, process):
+        self.process = process
 
         self.name = None
         self.author = None
@@ -681,14 +737,22 @@ class Engine(object):
         self.queue = queue.Queue()
         self.stdin_thread = threading.Thread(target=self._stdin_thread_target)
         self.stdin_thread.daemon = True
-        self.stdin_thread.start()
 
         self.return_code = None
         self.terminated = threading.Event()
 
         self.info_handlers = []
 
+        self.process.spawn(self)
+        self.stdin_thread.start()
+
+    def send_line(self, line):
+        LOGGER.debug("%s << %s", self.process, line)
+        return self.process.send_line(line)
+
     def on_line_received(self, buf):
+        LOGGER.debug("%s >> %s", self.process, buf)
+
         command_and_args = buf.split(None, 1)
         if not command_and_args:
             return
@@ -723,7 +787,7 @@ class Engine(object):
             if not self.is_alive():
                 break
 
-            command._execute(self)
+            command.execute(self)
             self.queue.task_done()
 
         self.on_terminated()
@@ -1002,11 +1066,19 @@ class Engine(object):
         option = Option(" ".join(name), type, default, min, max, var)
         self.options[option.name] = option
 
-    def queue_command(self, command):
+    def _queue_command(self, command, async_callback=None):
         if self.terminated.is_set():
             raise RuntimeError('can not queue command for terminated uci engine')
+
         self.queue.put(command)
-        return command._wait_or_callback()
+
+        if async_callback is True:
+            return command
+        elif async_callback:
+            command.add_done_callback(async_callback)
+            return command
+        else:
+            return command.result()
 
     def uci(self, async_callback=None):
         """
@@ -1017,7 +1089,7 @@ class Engine(object):
 
         :return: Nothing
         """
-        return self.queue_command(UciCommand(async_callback))
+        return self._queue_command(UciCommand(), async_callback)
 
     def debug(self, on, async_callback=None):
         """
@@ -1030,7 +1102,7 @@ class Engine(object):
 
         :return: Nothing
         """
-        return self.queue_command(DebugCommand(on, async_callback))
+        return self._queue_command(DebugCommand(on), async_callback)
 
     def isready(self, async_callback=None):
         """
@@ -1041,7 +1113,7 @@ class Engine(object):
 
         :return: Nothing
         """
-        return self.queue_command(IsReadyCommand(async_callback))
+        return self._queue_command(IsReadyCommand(), async_callback)
 
     def setoption(self, options, async_callback=None):
         """
@@ -1051,7 +1123,7 @@ class Engine(object):
 
         :return: Nothing
         """
-        return self.queue_command(SetOptionCommand(options, async_callback))
+        return self._queue_command(SetOptionCommand(options), async_callback)
 
     # TODO: Implement register command
 
@@ -1065,7 +1137,7 @@ class Engine(object):
 
         :return: Nothing
         """
-        return self.queue_command(UciNewGameCommand(async_callback))
+        return self._queue_command(UciNewGameCommand(), async_callback)
 
     def position(self, board, async_callback=None):
         """
@@ -1078,11 +1150,11 @@ class Engine(object):
         If the position is from a new game it is recommended to use the
         *ucinewgame* command before the *position* command.
 
-        :param board: A *chess.Bitboard*.
+        :param board: A *chess.Board*.
 
         :return: Nothing
         """
-        return self.queue_command(PositionCommand(board, async_callback))
+        return self._queue_command(PositionCommand(board), async_callback)
 
     def go(self, searchmoves=None, ponder=False, wtime=None, btime=None, winc=None, binc=None, movestogo=None, depth=None, nodes=None, mate=None, movetime=None, infinite=False, async_callback=None):
         """
@@ -1113,7 +1185,7 @@ class Engine(object):
             elements may be *None*. **In infinite search mode** there is no
             result. See *stop* instead.
         """
-        return self.queue_command(GoCommand(searchmoves, ponder, wtime, btime, winc, binc, movestogo, depth, nodes, mate, movetime, infinite, async_callback))
+        return self._queue_command(GoCommand(searchmoves, ponder, wtime, btime, winc, binc, movestogo, depth, nodes, mate, movetime, infinite), async_callback)
 
     def stop(self, async_callback=None):
         """
@@ -1123,7 +1195,7 @@ class Engine(object):
             *go* command. Results of infinite searches will also be available
             here.
         """
-        return self.queue_command(StopCommand(async_callback))
+        return self._queue_command(StopCommand(), async_callback)
 
     def ponderhit(self, async_callback=None):
         """
@@ -1134,7 +1206,7 @@ class Engine(object):
 
         :return: Nothing
         """
-        return self.queue_command(PonderhitCommand(async_callback))
+        return self._queue_command(PonderhitCommand(), async_callback)
 
     def quit(self, async_callback=None):
         """
@@ -1142,9 +1214,9 @@ class Engine(object):
 
         :return: The return code of the engine process.
         """
-        return self.queue_command(QuitCommand(self, async_callback))
+        return self._queue_command(QuitCommand(), async_callback)
 
-    def terminate(self, async=None):
+    def terminate(self, async=False):
         """
         Terminate the engine.
 
@@ -1157,11 +1229,11 @@ class Engine(object):
         self.process.close_std_streams()
         self.process.terminate()
 
-        promise = QuitCommand(self)
+        promise = TerminationPromise(self)
         if async:
             return promise
         else:
-            return promise.wait()
+            return promise.result()
 
     def kill(self, async=False):
         """
@@ -1174,19 +1246,16 @@ class Engine(object):
         self.process.close_std_streams()
         self.process.kill()
 
-        promise = QuitCommand(self)
+        promise = TerminationPromise(self)
         if async:
             return promise
         else:
-            return promise.wait()
+            return promise.result()
 
     def is_alive(self):
         """Poll the engine process to check if it is alive."""
         return self.process.is_alive()
 
-    def send_line(self, line):
-        return self.process.send_line(line)
-
 
 def popen_engine(command, engine_cls=Engine):
     """
@@ -1205,7 +1274,8 @@ def popen_engine(command, engine_cls=Engine):
     The input and input streams will be linebuffered and able both Windows
     and Unix newlines.
     """
-    return engine_cls(PopenProcess, (command, ))
+    process = PopenProcess(command)
+    return engine_cls(process)
 
 
 def spur_spawn_engine(shell, command, engine_cls=Engine):
@@ -1219,4 +1289,5 @@ def spur_spawn_engine(shell, command, engine_cls=Engine):
 
     .. _Spur: https://pypi.python.org/pypi/spur
     """
-    return engine_cls(SpurProcess, (shell, command))
+    process = SpurProcess(shell, command)
+    return engine_cls(process)
